/*
 * Copyright (c) 2005
 *    Christian Biere <christianbiere@gmx.de> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "connection.h"

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include "nettools.h"
#include "mem.h"
#include "oop.h"

static mem_cache_t *mc;

typedef struct {
  struct connection   vtable;
  uint64_t            magic;
  int                 fd;
  net_addr_t          addr;
  uint16_t            port;
  time_t              stamp;
  void                *context;
  void                (*event_cb)(connection_t *, connection_event_t);
  unsigned int        maxidle;
  time_t              established;
  connection_event_t  eventmask;
  bool                listening;
} connection_private_t;

static const uint64_t connection_magic =
  MAGIC_CHARS('C', 'O', 'N', 'N', 'C', 'T', 'O', 'N');
  
typedef union {
  connection_t  pub;
  connection_private_t priv;
} connection_union;

ATTRIBUTE_GET(connection, net_addr_t, addr)
ATTRIBUTE_GET(connection, uint16_t, port)
ATTRIBUTE_GET(connection, int, fd)
ATTRIBUTE_GET(connection, time_t, stamp)
  
ATTRIBUTE_GET(connection, void *, context)
ATTRIBUTE_SET(connection, void *, context)
ATTRIBUTE_GET(connection, unsigned int, maxidle)
ATTRIBUTE_SET(connection, unsigned int, maxidle)
ATTRIBUTE_SET(connection, connection_event_cb, event_cb)
ATTRIBUTE_GET(connection, connection_event_t, eventmask)
ATTRIBUTE_SET(connection, connection_event_t, eventmask)
ATTRIBUTE_GET(connection, bool, listening)

#define PRIV(obj) (((connection_union *)(obj))->priv)

static char *
connection_get_addrstr(connection_t *c, char *dst, size_t size)
{
  const net_addr_t addr = PRIV(c).addr;
  char *s;
  
  s = print_net_addr(dst, size, addr);
  assert(s != NULL && s != dst);
  return dst;
}

static void
connection_event(connection_t *c, connection_event_t ev, time_t now)
{
  connection_private_t *cp = &PRIV(c);
  
  cp->stamp = now;
  cp->event_cb(c, ev);
}

static void
connection_close(connection_t *c)
{
  assert(PRIV(c).fd >= 0);
#if 0 
  {
    char buf[NET_ADDR_PORT_BUFLEN];
    
    print_net_addr_port(buf, sizeof buf, PRIV(c).addr, PRIV(c).port);
    VERB("Closing connection to %s (fd=%d)", buf, PRIV(c).fd);
  }
#endif 
  if (close(PRIV(c).fd)) {
    CRIT("close(%d) failed: %s", PRIV(c).fd, compat_strerror(errno));
  }
      
  PRIV(c).fd = -1;
}

static connection_private_t *
connection_alloc(void)
{
  connection_private_t *c;

  if (!mc) {
    if (NULL == (mc = MEM_NEW(sizeof *c)))
      return NULL;
  }

  c = mem_alloc(mc);
  return c;
}

static void
connection_free(connection_private_t *c)
{
  mem_free(mc, c);
}


static void
connection_destruct(connection_t *c)
{
  connection_private_t *cp = &PRIV(c);
  
  memset(cp, 0, sizeof *cp);
  connection_free(cp);
}

static int64_t
connection_get_age(connection_t *c)
{
  connection_private_t *cp = &PRIV(c);
  return difftime(cp->stamp, cp->established);
}

static int
connection_set_buf(connection_t *c, int optname, int optval)
{
  int fd, ret;
  socklen_t optlen = sizeof optval;
  
  fd = PRIV(c).fd;
  ret = setsockopt(fd, SOL_SOCKET, optname, &optval, optlen);
  if (ret) {
    WARN("setsockopt(%d, SOL_SOCKET, %s, ...) failed: %s\n",
        fd, optname == SO_RCVBUF ? "SO_RCVBUF" : "SO_SNDBUF",
        compat_strerror(errno));
  }
  return ret;
}

static int
connection_set_rcvbuf(connection_t *c, int optval)
{
  return connection_set_buf(c, SO_RCVBUF, optval);
}

static int
connection_set_sndbuf(connection_t *c, int optval)
{
  return connection_set_buf(c, SO_SNDBUF, optval);
}

/**
 * Enables or disables the TCP_NODELAY socket option (Nagle algorithm) on
 * the connection socket.
 *
 * @param value if true, TCP_NODELAY is enabled, otherwise it's disabled.
 * @return -1 on failure (see errno) and zero on success.
 */
static int
connection_set_nodelay(connection_t *c, bool value)
{
  int val;

  assert(c);

  val = value ? 1 : 0;
  if (setsockopt(PRIV(c).fd, IPPROTO_TCP, TCP_NODELAY, &val, sizeof val)) {
    WARN("Cannot set TCP_NODELAY: %s", compat_strerror(errno));
    return -1;
  }
  return 0;
}

static int
connection_set_tos_lowdelay(connection_t *c)
{
  int val = IPTOS_LOWDELAY;
  
  if (setsockopt(PRIV(c).fd, IPPROTO_IP, IP_TOS, &val, sizeof val)) {
    WARN("Cannot set IPTOS_LOWDELAY: %s", compat_strerror(errno));
    return -1;
  }

  return 0;
}

static int
connection_set_tos_throughput(connection_t *c)
{
  int val = IPTOS_THROUGHPUT;
  
  if (setsockopt(PRIV(c).fd, IPPROTO_IP, IP_TOS, &val, sizeof val)) {
    WARN("Cannot set IPTOS_THROUGHPUT: %s", compat_strerror(errno));
    return -1;
  }

  return 0;
}

static int
connection_set_tos_reliability(connection_t *c)
{
  int val = IPTOS_RELIABILITY;
  
  if (setsockopt(PRIV(c).fd, IPPROTO_IP, IP_TOS, &val, sizeof val)) {
    WARN("Cannot set IPTOS_RELIABILITY: %s", compat_strerror(errno));
    return -1;
  }

  return 0;
}

static int
connection_set_rcvlowat(connection_t *c, unsigned int lowat)
{
#ifdef SO_RCVLOWAT
  if (setsockopt(PRIV(c).fd, SOL_SOCKET, SO_RCVLOWAT, &lowat, sizeof lowat)) {
    WARN("Cannot set SO_RCVLOWAT: %s", compat_strerror(errno));
    return -1;
  }
#else
  (void) c;
  (void) lowat;
#endif

  return 0;
}

static int
connection_set_blocking(connection_t *c, bool block)
{
  int flags;

  flags = fcntl(PRIV(c).fd, F_GETFL);
  if (-1 == flags) {
    CRIT("fcntl(fd, F_GETFL) failed: %s", compat_strerror(errno));
    return -1;
  }

  if (block) {
    if (0 == (flags & O_NONBLOCK))
      return 0;
    flags &= ~O_NONBLOCK;
  } else {
    if (0 != (flags & O_NONBLOCK))
      return 0;
    flags |= O_NONBLOCK;
  }
  
  if (-1 == fcntl(PRIV(c).fd, F_SETFL, flags)) {
    CRIT("fcntl(fd, F_SETFL) failed: %s", compat_strerror(errno));
    return -1;
  }
  return 0;
}


static const struct connection connection_vtable = {
  &connection_get_addr,
  &connection_get_port,
  &connection_get_fd,
  &connection_get_stamp,
  &connection_get_addrstr,
  &connection_event,
  &connection_set_event_cb,
  &connection_get_eventmask,
  &connection_set_eventmask,
  &connection_get_context,
  &connection_set_context,
  &connection_get_maxidle,
  &connection_set_maxidle,
  &connection_get_age,
  &connection_set_rcvbuf,
  &connection_set_sndbuf,
  &connection_set_nodelay,
  &connection_set_tos_lowdelay,
  &connection_set_tos_throughput,
  &connection_set_tos_reliability,
  &connection_set_blocking,
  &connection_set_rcvlowat,
  &connection_get_listening,
  &connection_close,
  &connection_destruct
};

connection_t * 
connection_listen(const net_addr_t addr, uint16_t port, int backlog)
{
  const struct sockaddr *sa;
  connection_t *c;
  int reuse_addr;
  socklen_t len;
  int fd;

  fd = socket(net_addr_family(addr), SOCK_STREAM, 0);
  if (fd < 0) {
    CRIT("socket() failed: %s", compat_strerror(errno));
    return NULL;
  }

  reuse_addr = 1;
  len = sizeof reuse_addr;
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, len)) {
    CRIT("setsockopt() failed: %s", compat_strerror(errno));
    close(fd);
    return NULL;
  }

  len = net_addr_sockaddr(addr, port, &sa);
  if (bind(fd, sa, len)) {
    CRIT("bind() failed: %s", compat_strerror(errno));
    close(fd);
    return NULL;
  }

  if (listen(fd, backlog)) {
    CRIT("listen() failed: %s", compat_strerror(errno));
    close(fd);
    return NULL;
  }

  c = connection_new(fd, addr, port, compat_mono_time(NULL));
  if (c) {
    PRIV(c).listening = true;
  }
  return c;
}

connection_t * 
connection_udp(const net_addr_t addr, uint16_t port)
{
  const struct sockaddr *sa;
  connection_t *c;
  socklen_t len;
  int fd;

  fd = socket(net_addr_family(addr), SOCK_DGRAM, 0);
  if (fd < 0) {
    CRIT("socket() failed: %s", compat_strerror(errno));
    return NULL;
  }

  len = net_addr_sockaddr(addr, port, &sa);
  if (bind(fd, sa, len)) {
    int e = errno;
    
    CRIT("bind() failed: %s", compat_strerror(errno));
    close(fd);
    errno = e;
    return NULL;
  }

  c = connection_new(fd, addr, port, compat_mono_time(NULL));
  if (!c) {
    close(fd);
  }
  return c;
}

connection_t *
connection_accept(connection_t *server)
{
  static bool inherits = false;
#ifdef HAVE_IPV6_SUPPORT
  struct sockaddr_storage sa;
#else
  struct sockaddr sa;
#endif
  socklen_t len = sizeof sa;
  net_addr_t addr;
  connection_t *c;
  int fd, s_fd;
  uint16_t port;

  s_fd = server->get_fd(server);
  fd = accept(s_fd, cast_to_void(&sa), &len);
  if (fd < 0) {
    if (errno != ECONNABORTED)
      CRIT("accept() failed: %s", compat_strerror(errno));
    return NULL;
  }

#ifdef HAVE_IPV6_SUPPORT
  switch (sa.ss_family) {
  case PF_INET6:
    {
      const struct sockaddr_in6 *sin6 = cast_to_void(&sa);
      
      addr = net_addr_set_ipv6(cast_to_const_void(&sin6->sin6_addr));
      port = ntohs(sin6->sin6_port);
    }
    break;
#else
  switch (sa.sa_family) {
#endif

  case PF_INET:
    {
      const struct sockaddr_in *sin = cast_to_void(&sa);
      
      addr = net_addr_set_ipv4(sin->sin_addr.s_addr);
      port = ntohs(sin->sin_port);
    }
    break;

  default:
    {
      unsigned u =
#ifdef HAVE_IPV6_SUPPORT
        sa.ss_family;
#else
        sa.sa_family;
#endif

      WARN("accept()ed connection of unknown protocol (%u), closing", u);
    }
    close(fd);
    return NULL;
  }

  c = connection_new(fd, addr, port, server->get_stamp(server));

  if (!inherits) {
    static bool checked = false;
    int c_flags, s_flags;
    
    c_flags = fcntl(fd, F_GETFL);
    if (c_flags == -1) {
      CRIT("fcntl(fd, F_GETFL) failed: %s", compat_strerror(errno));
      c_flags = 0;
    }
    
    s_flags = fcntl(s_fd, F_GETFL);
    if (s_flags == -1) {
      CRIT("fcntl(fd, F_GETFL) failed: %s", compat_strerror(errno));
    }

    if (s_flags != -1 && !checked) {
      checked = true;     
      inherits = (s_flags & O_NONBLOCK) == (c_flags & O_NONBLOCK);
      DBUG("client socket %s O_NONBLOCK from server",
        inherits ? "inherits" : "does NOT inherit");
    }

    c_flags = (c_flags & ~O_NONBLOCK) | (s_flags & O_NONBLOCK);
    if (-1 == fcntl(fd, F_SETFL, c_flags)) {
      CRIT("fcntl(fd, F_SETFL) failed: %s", compat_strerror(errno));
    }
  }
  
 #if 0
  {
    int ret;
    int optval;
    socklen_t optlen = sizeof optval;

    ret = getsockopt(fd, SOL_SOCKET, SO_SNDLOWAT, &optval, &optlen);
    if (ret)
      s->public.error(&s->public, __func__, "getsockopt failed: ", errno);
    printf("SO_SNDLOWAT: %d\n", optval);
  
    ret = getsockopt(fd, SOL_SOCKET, SO_RCVLOWAT, &optval, &optlen);
    if (ret)
      s->public.error(&s->public, __func__, "getsockopt failed: ", errno);
    printf("SO_RCVLOWAT: %d\n", optval);
  }
#endif /* 0 */
  
  if (!c) {
    CRIT("connection_new() failed");
    close(fd);
  }
  return c;
}

connection_t *
connection_new(int fd, const net_addr_t addr, uint16_t port, time_t now)
{
  connection_private_t *c;
  
  c = connection_alloc();
  if (c) {
    c->vtable = connection_vtable;
    c->magic = connection_magic;
    c->fd = fd;
    c->addr = addr;
    c->port = port;
    c->stamp = now;
    c->established = now;
    c->context = NULL;
    c->event_cb = NULL;
    c->maxidle = 0;
    c->eventmask = CONN_EV_NONE;
    c->listening = false;
#if 0
    int optval;
    socklen_t optlen = sizeof optval;

    ret = getsockopt(fd, SOL_SOCKET, SO_SNDBUF, &optval, &optlen);
    if (!ret) {
      DBUG("SO_SNDBUF: %d\n", optval);
    }
  
    ret = getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &optval, &optlen);
    if (!ret) {
      DBUG("SO_RCVBUF: %d\n", optval);
    }
#endif /* 0 */
  }
  return (connection_t *) c;
}

/* vi: set ai et sts=2 sw=2 cindent: */
