/*
 * Copyright (c) 2005
 *    Christian Biere <christianbiere@gmx.de> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "mem.h"

union mem_chunk {
  void      *next;
  uint8_t   u8;
  uint16_t  u16;
  uint32_t  u32;
  uint64_t  u64;
  float     f;
  double    d;
};

struct mem_pool {
	union mem_chunk chunks[1];
};

struct mem_cache {
  const char *name;
	struct mem_pool **pools;
	union mem_chunk *avail;
  size_t num_pools;
  size_t size, hold;
};

struct mem_pool *
mem_pool_new(size_t size, size_t hold)
{
  struct mem_pool *mp;
  
  assert(size > 0);
  assert(hold > 0);
  
  mp = calloc(1, hold * size * sizeof mp->chunks[0]);
  if (mp) {
    size_t i, j;
    
    for (i = 0, j = 0; i < hold - 1; i++, j += size)
      mp->chunks[j].next = &mp->chunks[j + size];

    mp->chunks[j].next = NULL;
  }

  return mp;
}

mem_cache_t *
mem_new(size_t size, const char *name)
{
  mem_cache_t *mc;

  assert(size > 0);
  
  mc = calloc(1, sizeof *mc);
  if (mc) {
    mc->name = name ? name : "(null)";

    mc->size = size / sizeof(union mem_chunk) +
      (size % sizeof(union mem_chunk) ? 1 : 0);
    mc->hold = 8 * MAX(1,
        compat_getpagesize() / (mc->size * sizeof(union mem_chunk)));
    mc->num_pools = 0;
    mc->avail = NULL;
    mc->pools = NULL;

    DBUG("mem_new(): name=%s hold=%d, size=%d",
      mc->name, (int) mc->hold, (int) mc->size);
  }
  return mc;
}
 
void *
mem_alloc(mem_cache_t *mc)
{
  void *p;

  assert(mc);
  
  if (NULL != (p = mc->avail)) {
    mc->avail = mc->avail->next;
  } else {
    struct mem_pool *mp;
    
    mp = mem_pool_new(mc->size, mc->hold);
    if (mp) {
      void *q;
      
      q = realloc(mc->pools, (1 + mc->num_pools) * sizeof mc->pools[0]);
      if (q) {
        mc->pools = q;
        mc->pools[mc->num_pools++] = mp;
      
        mc->avail = &mp->chunks[0];
        p = mc->avail;
        mc->avail = mc->avail->next;

        memset(p, 0, mc->size * sizeof mp->chunks[0]);
      } else {
        free(mp);
      }
    }
  }

  return p;
}

void
mem_free(mem_cache_t *mc, void *p)
{
  assert(mc);

  if (p) {
    union mem_chunk *c = p;
   
    c->next = mc->avail;
    mc->avail = c;
  }
}

/**
 * Frees the mem_cache_t structure and all cached freed chunks. It DOES NOT
 * free chunks still in use.
 */
void
mem_destruct(mem_cache_t *mc)
{
  if (mc) {
    size_t i;

    for (i = 0; i < mc->num_pools; i++) {
      free(mc->pools[i]);
    }
    free(mc->pools);
    free(mc);
  }
}

/* vi: set ai et ts=2 sts=2 sw=2 cindent: */
